/* 
    xWinTV using Bt848 frame grabber driver

    Copyright (C) 1998 Moritz Wenk (wenk@mathematik.uni-kl.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// Hey, better do not change the order of the includes !!!
// REMEMBER: Qt-stuff BEFORE X11-stuff !!!

#include <wintv.h>

#include <stdlib.h>
#include <iostream.h>
//#include <iomanip.h>
#include <fstream.h>
#include <strstream.h>
#include <unistd.h>
//#include <linux/string.h> // for memcpy  /* glibc */
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/errno.h> // for RedHat

#include "wintvscreen.h"

//#include <kapp.h>
#include <kiconloader.h>
#include <qpixmap.h>
//#include <qcolor.h>
#include <qpainter.h> 

// video clip stuff
#include "avivideoclip.h"
#include "rawvideoclip.h"
#include "ppmvideoclip.h"

// includes needed for X stuff 
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/IntrinsicP.h>
#include <X11/ObjectP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xmd.h>

#ifdef HAVE_LIBXXF86DGA
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86dgastr.h>
#endif

#include "qt_visual.h"

// debug switches
#ifdef myDEBUG
#define tvDEBUG
//#define X11DEBUG
#endif


#ifdef X11DEBUG
#include "xeventtype.h"
#endif

//#undef HAVE_LIBJPEG
#ifdef HAVE_LIBJPEG
extern "C" {          // jpeglib is pure C 
#include <jpeglib.h> // use KDE jpeglib.h
}
#endif

// FIXME

#define HEX(N) hex << setw(N) << setfill(int('0')) 
#define DEC(N) dec << setw(N) << setfill(int('0')) 

int one=1;
int zero=0;

extern int errno;  // for RedHat
XErrorHandler oldXErrorHandler;

//==============================================================================
//==============================================================================
// einige X errors muessen (sollten) ignoriert werden,
// z.b. BadWindow und so.
// Qt wuerde sonst mit fatal(...) aussteigen...
//
XErrorHandler xerror(Display *, XErrorEvent *event)
{
  switch (event->error_code){
  case BadWindow:
#ifdef tvDEBUG
    warning("screen: X error BadWindow ingnored");
#endif
    return 0 ;
    break;
      
  case BadMatch:
#ifdef tvDEBUG
    warning("screen: X error BadMatch ingnored");
#endif
    return 0 ;
    break;

  case BadColor:
#ifdef tvDEBUG
    warning("screen: X error BadColor ingnored");
#endif
    return 0 ;
    break;

  default:
#ifdef tvDEBUG
    warning("screen: unknown X error %d ingnored",event->error_code);
#endif
    return 0;
  }
}

//============================= winTVScreenWidget =================================================
//============================= winTVScreenWidget =================================================

// local variables to avoid inclusion of X11 header stuff in wintvscreen.h
XColor xcolors[256];
Colormap xcolormap;       // colormap, needed when having only 256 colors -> pseudocolor

winTVScreenWidget::winTVScreenWidget( QWidget *_parent, 
				      const char *name,
				      int _capture, 
				      const char *_bttvdev, 
				      int _bpp,int _dual,
				      int _freq, 
				      int _input, 
				      char *_visualname 
				      )
  : QWidget( _parent, name ), 
    devname(_bttvdev),
    captureIsOn(_capture), 
    defaultFreq(_freq), 
    defaultInput(_input), 
    visname(_visualname),
    parent(_parent),
    dual(_dual),
    bpp(_bpp)
{ 
#ifdef tvDEBUG
  debug("screen: begin, captureIsOn %d",captureIsOn);
#endif

    // do some widget stuff
    //setCaption("xWinTV - TV Screen");
    //    setBackgroundcolor
    setMinimumSize( 10, 10 );
    setMaximumSize(32000,32000);
    //resize(win_default_width(),win_default_height());

    QColor bkgc(0,0,0);
    setBackgroundColor(bkgc); 
    
    KIconLoader *loader = kapp->getIconLoader();
    title_pm = loader->loadIcon("kwintv_titel.xpm");
    
    // set background pixmap
    //setBackgroundPixmap(title_pm);

    /////////////////////////////////////////////////////////////////////////
    // channel info
    chInfoLCD = new QLCDNumber( NULL, "lcd");
    QPoint p(0,0);
    chInfoLCD->recreate( NULL, WStyle_Customize | WStyle_NoBorder | WStyle_Tool, p );
    chInfoLCD->setSmallDecimalPoint( false );
    chInfoLCD->setNumDigits( 3 );
    chInfoLCD->setMode( QLCDNumber::DEC );
    chInfoLCD->setSegmentStyle( QLCDNumber::Flat );
    chInfoLCD->setFrameStyle( QFrame::NoFrame );
    chInfoLCD->hide();

    chInfoText = new QLabel( NULL, "label", WStyle_Customize | WStyle_NoBorder | WStyle_Tool );
    chInfoText->setFrameStyle( QFrame::NoFrame );
    chInfoText->setAutoResize(true);
    chInfoText->hide();

    chInfoBar = new KProgress (0, 255, 0, KProgress::Horizontal, NULL, "bar");
    chInfoBar->setBarStyle(KProgress::Blocked);
    chInfoBar->recreate( NULL, WStyle_Customize | WStyle_NoBorder | WStyle_Tool, p );
    chInfoBar->setFrameStyle( QFrame::NoFrame );
    chInfoBar->hide();

    {
      QFont f("Helvetica", 48, QFont::Bold);
      chInfoText->setFont( f );
      QColorGroup normal( QColor( QRgb(0) ), QColor( QRgb(0) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(255) ), QColor( QRgb(0) ) );
      QColorGroup disabled( QColor( QRgb(8421504) ), QColor( QRgb(12632256) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(8421504) ), QColor( QRgb(12632256) ) );
      QColorGroup active( QColor( QRgb(0) ), QColor( QRgb(12632256) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(0) ), QColor( QRgb(16777215) ) );
      QPalette palette( normal, disabled, active );
      chInfoText->setPalette( palette );
    }
    {
      QColorGroup normal( QColor( QRgb(14319217) ), QColor( QRgb(0) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(0) ), QColor( QRgb(16777215) ) );
      QColorGroup disabled( QColor( QRgb(8421504) ), QColor( QRgb(12632256) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(8421504) ), QColor( QRgb(12632256) ) );
      QColorGroup active( QColor( QRgb(0) ), QColor( QRgb(12632256) ), QColor( QRgb(16777215) ), QColor( QRgb(6316128) ), QColor( QRgb(10789024) ), QColor( QRgb(0) ), QColor( QRgb(16777215) ) );
      QPalette palette( normal, disabled, active );
      chInfoLCD->setPalette( palette );
      chInfoBar->setPalette( palette );
      //chInfoText->setPalette( palette );
    }

    chInfo=AKTIVE;
    //chInfo=INAKTIVE; 
    infoTimer = new QTimer(this);
    connect(infoTimer,SIGNAL(timeout()),this,SLOT(unShowChannelInfoSLOT()));

    /////////////////////////////////////////////////////////////////////////
    // timer for view refresh actions
    viewRefreshTimer= new QTimer(this);
    connect(viewRefreshTimer,SIGNAL(timeout()),this,SLOT(refreshTVActionSLOT()));

    /////////////////////////////////////////////////////////////////////////
    // timer for autodetection of audio mode
    audioDetectTimer = new QTimer(this);
    connect(audioDetectTimer,SIGNAL(timeout()),this,SLOT(autoDetectAudioModeSLOT()));

    // init everything (bttv)
#ifdef tvDEBUG
    debug("screen: initialize");
#endif 

    /////////////////////////////////////////////////////////////////////////
    // set up device
    if ( -1 == (fbttv=open(devname,O_RDWR)) ) {
      fatal("kwintv: Error opening the Video4Linux device <%s>!",devname);
    }

    // capture off
    if ( -1 == ioctl(fbttv, VIDIOCCAPTURE, &zero) )
      fatal("screen: VIDIOCCAPTURE in ::winTVScreenWidget (off)");

    /////////////////////////////////////////////////////////////////////////
    // use my XErrorHandler 
    oldXErrorHandler= XSetErrorHandler((XErrorHandler) xerror);

    initializeTV(parent->winId());

    /////////////////////////////////////////////////////////////////////////
    // final init of bttv
    // unmute, audio mode = auto
    /*
    vaudio[0].flags &= ~VIDEO_AUDIO_MUTE;
    if (-1 == ioctl(fbttv,VIDIOCSAUDIO,&vaudio[0]))
      fatal("ioctl VIDIOCSAUDIO");
      */

    /////////////////////////////////////////////////////////////////////////
    // install filter for special X events
    installEventFilter( this );

    // we are running
    widget_is_active=1;
    captureWasOnQt=captureIsOn;
    captureWasOnX11=captureIsOn;

#ifdef tvDEBUG
  debug("screen: end");
#endif
}

//==============================================================================

void winTVScreenWidget::postInit()
{
  if ( captureIsOn ) {
    newTVGeometry( false );
    captureState=1;
    if ( -1 == ioctl(fbttv, VIDIOCCAPTURE, &one) )
      fatal("screen: VIDIOCCAPTURE in ::postInit (on)");
  } else {
    if( -1 == ioctl(fbttv, VIDIOCCAPTURE, &zero) )
      fatal("screen: VIDIOCCAPTURE in ::postInit (off)");
  }
}

//==============================================================================

winTVScreenWidget::~winTVScreenWidget()
{
  if (widget_is_active) {
    widget_is_active=0;
    destroyTV();
  } 
#ifdef tvDEBUG
  else 
    debug("screen: winTVScreenWidget already destroyed");
#endif
}

//==============================================================================
// X11 events des shell windows

void winTVScreenWidget::Shellx11EventSLOT( XEvent *event )
{
  switch_capture(0);

#ifdef X11DEBUG
  debug("X11: x11Event: shell event %s",events[event->type]);
#endif

  switch ( event->type ) {
  case MapNotify:
    captureIsOn=captureWasOnX11;
    getTVClip();
    break;
  case UnmapNotify:
    captureWasOnX11=captureIsOn;
    captureIsOn=false;
    if ( allowViewRefresh ) {
      refreshTV();
    }
    break;
  } // switch
  
  switch_capture(1);
}

//==============================================================================
// X11 events des TV windows

bool winTVScreenWidget::x11Event( XEvent *event )
{
  Window wwin;

  if ((!isVisible())||((wwin=(Window) winId()) == None) || (event->xvisibility.window != wwin)) 
    return(false);

  if (event->type == VisibilityNotify) {
#ifdef X11DEBUG
    debug("X11: x11Event: view event %s",events[event->type]);
#endif
    switch_capture(0);

    switch (event->xvisibility.state) {
    case VisibilityFullyObscured:
      // switch capture off
      captureWasOnQt= captureIsOn;
      captureIsOn=false;
      break;
    case VisibilityPartiallyObscured:
#ifdef X11DEBUG
      debug("X11: x11Event: view event VisibilityPartiallyObscured");
#endif
      captureIsOn= captureWasOnQt;
      if ( captureIsOn ) {
	getTVClip();
	if ( allowViewRefresh && !didViewRefresh && allowViewRefreshVisibility ) {
	  refreshTV();
	} else didViewRefresh=false;
      } 
      break;
    case VisibilityUnobscured: 
#ifdef X11DEBUG
      debug("X11: x11Event: view event VisibilityUnobscured");
#endif
      captureIsOn= captureWasOnQt;
      didViewRefresh=false;
      if ( captureIsOn ) {
	getTVClip();
      }
      break;
    } // switch (event->xvisibility.state)
    switch_capture(1);     
  } // if (event->type)

  return (false);
}

//==============================================================================
// X11 events des root windows (desktop)

bool winTVScreenWidget::eventFilter( QObject *, QEvent *e )
{
  Window wwin;

  if ((!isVisible()) || (wwin=(Window) winId()) == None) return(false);

  if ( e->type() == XRWEvent ) {
    
    X11RWEvent *re = (X11RWEvent*)e;

    switch (re->XEventTyp()) {
    case MapNotify:
    case UnmapNotify:
    case ConfigureNotify:
#ifdef X11DEBUG
    debug("X11: eventFilter: root event %s",events[re->XEventTyp()]);
#endif
      if ( captureIsOn ) {
	switch_capture(0);
	getTVClip();
	switch_capture(1);
	captureWasOnQt= captureIsOn;
      } 
      break;
    } // switch

    return true;                        // eat event
  }

  return false;                         // standard event processing
}

//==============================================================================

void winTVScreenWidget::moveEvent ( QMoveEvent * theEvent )
{
  switch_capture(0);
  unShowChannelInfoSLOT();
  newTVGeometry( true );
  switch_capture(1);

#ifdef tvDEBUG
  debug("screen: move: %d, %d",theEvent->pos().x(),theEvent->pos().y());
#endif
}

//==============================================================================

void winTVScreenWidget::resizeEvent( QResizeEvent * theEvent )
{
  switch_capture(0);
  unShowChannelInfoSLOT();
  newTVGeometry( true );
  switch_capture(1);

#ifdef tvDEBUG
  debug("screen: resize: %d, %d",theEvent->size().width(),theEvent->size().height());
#endif
}

//==============================================================================

void winTVScreenWidget::paintEvent( QPaintEvent * event)
{
  if ( !captureIsOn ) {
    if ( !title_shown ) {
      QPainter pa(this);
      pa.setClipRect(event->rect());
      pa.drawPixmap((width()/2)-(title_pm.width()/2),(height()/2)-(title_pm.height()/2),title_pm);
    } else {
      //title_shown=1;
    }
  }
}

//==============================================================================

void winTVScreenWidget::keyPressEvent ( QKeyEvent *keyE )
{
  emit keyPressedSIGNAL(keyE->key());
  keyE->ignore();
}

//==============================================================================
/*
void winTVScreenWidget::closeEvent (  QCloseEvent * e )
{
#ifdef tvDEBUG
  debug("screen: closeEvent");
#endif
  destroyTV();
  e->accept();
}
*/
//==============================================================================
/*
void winTVScreenWidget::mouseMoveEvent(QMouseEvent *me)
{
#ifdef tvDEBUG
    debug("screen: mouseMoveEvent");
#endif

  if (cursorState) {
    if (mtimer->isActive()) mtimer->stop();
    mtimer->start(MOUSE_DELAY,true);
#ifdef tvDEBUG
    debug("screen: timer started");
#endif
  } else {
    didViewChangesCursorStateSLOT();
  }
		     
}
*/
// =============================================================================== 

void winTVScreenWidget::mouseReleaseEvent ( QMouseEvent * me)
{
#ifdef tvDEBUG
  debug("screen: mouseReleaseEvent");
#endif
  if ((me->button() == RightButton) && (tvPopup != NULL)) {
    tvPopup->popup(mapToGlobal(me->pos()));
  }
  if (me->button() == MidButton) {
    emit toggleTVonlySIGNAL();
  }
}


//==============================================================================

void winTVScreenWidget::switch_capture(int on)
{
#if defined X11DEBUG || defined tvDEBUG
  debug("screen: switch_capture: captureIsOn = %s, on = %s",captureIsOn?"on":"off",on?"on":"off");
#endif
  if (on && captureIsOn) {
    if ( isVisible() ) 
      if( -1 == ioctl(fbttv, VIDIOCCAPTURE, &one))
	fatal("screen: VIDIOCCAPTURE in ::switch_capture (on)");
  } else {
    if( -1 == ioctl(fbttv, VIDIOCCAPTURE, &zero))
      fatal("screen: VIDIOCCAPTURE in ::switch_capture (off)");
  }
}

//==============================================================================

void winTVScreenWidget::capture(int on)
{
#if defined X11DEBUG || defined tvDEBUG
  debug("screen: capture=%s",on?"on":"off");
#endif
  if (isVisible()) {
    if (on) {
      captureIsOn=true; 
      captureWasOnQt=captureIsOn;
      captureWasOnX11=captureIsOn;
      newTVGeometry( false );
      if ( -1 == ioctl(fbttv, VIDIOCCAPTURE, &one) )
	fatal("screen: VIDIOCCAPTURE in ::capture (on)");
    } else {
      if ( -1 == ioctl(fbttv, VIDIOCCAPTURE, &zero) )
	fatal("screen: VIDIOCCAPTURE in ::capture (off)");
      captureIsOn=false;
      captureWasOnQt=captureIsOn;
      captureWasOnX11=captureIsOn;
    }
  }
}
//==============================================================================

//==============================================================================

int winTVScreenWidget::initializeTV(WId parentWid)
{
  Window rootw;
  //int col;
  // uint rwidth;  // why unit ??
  int rwidth;
  int visual_class;
  int major, minor, bank, ram;
  Display *disp;
  ushort i;
  Visual *myvisual;
  //  XSetWindowAttributes swinattr;

  // get capabilities
  if( -1 == ioctl(fbttv,VIDIOCGCAP,&vcap) )
    fatal("screen: capabilitys");
#ifdef tvDEBUG
  debug("\t name: %s",vcap.name);
  for (i = 0; device_cap[i] != NULL; i++)
    if (vcap.type & (1 << i))
      debug("\t %s",device_cap[i]);
  debug("screen: range of tv size : %dx%d => %dx%d",
	vcap.minwidth,vcap.minheight,
	vcap.maxwidth,vcap.maxheight);
#endif  
  
  // get input channels
#ifdef tvDEBUG
  debug("screen: input channels: %d",vcap.channels);
#endif  
  vchan = (struct video_channel*)malloc(sizeof(struct video_channel)*vcap.channels);
  memset(vchan,0,sizeof(struct video_channel)*vcap.channels);
  
  for (i = 0; i < vcap.channels; i++) {
    vchan[i].channel = i; 
    if ( -1 == ioctl(fbttv,VIDIOCGCHAN,&vchan[i]) )
      fatal("screen: get input channels");
#ifdef tvDEBUG
    debug("\t %s: %d %s%s %s%s",
	  vchan[i].name,
	  vchan[i].tuners,
	  (vchan[i].flags & VIDEO_VC_TUNER)   ? "tuner "  : "",
	  (vchan[i].flags & VIDEO_VC_AUDIO)   ? "audio "  : "",
	  (vchan[i].type & VIDEO_TYPE_TV)     ? "tv "     : "",
	  (vchan[i].type & VIDEO_TYPE_CAMERA) ? "camera " : "");
#endif
  }

  /* ioctl probe, switch to input 0 */
  if ( -1 == ioctl(fbttv,VIDIOCSCHAN,&vchan[0]) ) {
    fatal("sreen: you need a newer bttv version (>= 0.5.14)\n");
  }
  
  // get audios 
#ifdef tvDEBUG
  debug("screen: input audios: %d",vcap.audios);
#endif

  vaudio = (struct video_audio*)malloc(sizeof(struct video_audio)*vcap.audios);
  memset(vaudio,0,sizeof(struct video_audio)*vcap.audios);
  for (i = 0; i < vcap.audios; i++) {
    vaudio[i].audio = i;
    if ( -1 == ioctl(fbttv,VIDIOCGAUDIO,&vaudio[i]) )
      warning("ioctl VIDIOCGAUDIO");
    //fatal("ioctl VIDIOCGAUDIO");
#ifdef tvDEBUG
    debug("\t %d (%s): ",i,vaudio[i].name);
    if (vaudio[i].flags & VIDEO_AUDIO_MUTABLE)
      debug("\t mutable, muted=%s ",(vaudio[i].flags&VIDEO_AUDIO_MUTE) ? "yes":"no");
    if (vaudio[i].flags & VIDEO_AUDIO_VOLUME)
      debug("\t volume=%d ",vaudio[i].volume);
    if (vaudio[i].flags & VIDEO_AUDIO_BASS)
      debug("\t bass=%d ",vaudio[i].bass);
    if (vaudio[i].flags & VIDEO_AUDIO_TREBLE)
      debug("\t treble=%d ",vaudio[i].treble);
    debug("\t mode=%d",vaudio[i].mode);
    if (vaudio[i].flags & VIDEO_AUDIO_VOLUME) debug("\t can volume, %d",vaudio[i].volume);
#endif
  }
  audioVolumeCan=(vaudio[0].flags & VIDEO_AUDIO_VOLUME);
  audioMode=vaudio[0].mode;
  //audioVolume=vaudio[0].volume;
  audioVolume=127;
  
  // check tuner
  if( -1 == ioctl(fbttv, VIDIOCGTUNER, &vtuner) )
    //fatal("screen: get tuner");
    warning("screen: VIDIOCGTUNER");
#ifdef tvDEBUG
  debug("screen: tuner: %s %lu-%lu",vtuner.name,vtuner.rangelow,vtuner.rangehigh);
  debug("screen: tuner can stereo: %s",vtuner.flags & VIDEO_TUNER_STEREO_ON?"yes":"no");
  debug("screen: tuner: supported norms: ");
  for (i = 0; norms[i].str != NULL; i++) {
    if (vtuner.flags & (1<<i)) {
      debug("\t %s",norms[i].str);
    }
  }
#endif

  // get window parameters
  if ( ioctl(fbttv, VIDIOCGWIN, &(vwin)) == -1 )
    fatal("screen: get window");
  // get frame buffer parameters
  if ( ioctl(fbttv, VIDIOCGFBUF, &(vbuf)) == -1 )
    fatal("screen: get frame buffer");

  rwidth=vbuf.width;

  // get the display
  disp=qt_xdisplay();

  // do some strange things with visuals and dga
  int flags;
  void *base = NULL;

#ifdef HAVE_LIBXXF86DGA
  bool have_dga= false;
  if (XF86DGAQueryExtension(disp, &major, &minor)) {
    XF86DGAQueryDirectVideo(disp,XDefaultScreen(disp),&flags);
    if (flags & XF86DGADirectPresent) {
      have_dga= true;
      XF86DGAGetVideoLL(disp, DefaultScreen(disp),(int *)&base,
			&rwidth, &bank, &ram);
      XF86DGAQueryVersion(dpy,&major,&minor);
      if ((major != XF86DGA_MAJOR_VERSION) || (minor != XF86DGA_MINOR_VERSION)) {
	warning("screen: X-Server DGA extension version mismatch, disabled");
	warning("screen: server version %d.%d != included version %d.%d",major,minor,
		XF86DGA_MAJOR_VERSION,XF86DGA_MINOR_VERSION);
	have_dga= false;
      }
      if (!base) fatal("screen: can not allocate video buffer base adress: 0x%x",(int)base);
    }
  }
#ifdef tvDEBUG
      debug("screen: X-Server supports DGA extention (version %d.%d)",major,minor);
#endif
#endif /* HAVE_LIBXXF86DGA */

  // FIX ME visuals
  // get the visual and check for doublebuffering
  if (!dual) {
    dual=find_visual(disp,visname,&myvisual,&visual_class);
  }
  // FIX MEX
  if (dual) vbuf.depth=32;

#ifdef HAVE_LIBXXF86DGA
  if (have_dga) {
    if ( (void*)((unsigned int)vbuf.base & 0xffff0000) != base ) {  /* XXX matrox of-by-1024 hack */
      warning("v4l and DGA extension disagree about the framebuffer address: v4l: 0x%x, dga: 0x%x (needed).",
	      (unsigned int)vbuf.base & 0xffff0000,(unsigned int)base);
      vbuf.base=(void *)base;
      if ( -1 == ioctl(fbttv, VIDIOCSFBUF, &vbuf) ) 
	fatal("must be root to set frame buffer");
    }
    if ( bpp ) {
      vbuf.depth= bpp;
      if ( -1 == ioctl(fbttv, VIDIOCSFBUF, &vbuf) ) 
	fatal("must be root to set colour depth");
    } else 
      if ( vbuf.depth != QColor::numBitPlanes() ) {
	warning("v4l and DGA extension disagree about the colour depth: %d bpp, need %d bpp.",
		vbuf.depth,QColor::numBitPlanes());
	vbuf.depth=QColor::numBitPlanes();
	if ( -1 == ioctl(fbttv, VIDIOCSFBUF, &vbuf) ) 
	  fatal("must be root to set colour depth");
      }
    // FIX ME Test auf richtige screen size
    // does this work???
    if ( vbuf.bytesperline != rwidth*vbuf.depth/8 ) {
      int val=vbuf.depth;
      if(val==15) val++;
      if ( vbuf.bytesperline != rwidth*val/8 ) {
	warning("v4l and DGA extension disagree about the number of bytes per line: %d, need %d.",
		vbuf.bytesperline,rwidth*val/8);
	vbuf.bytesperline=rwidth*val/8;
	if( -1 == ioctl(fbttv, VIDIOCSFBUF, &vbuf)) 
	  fatal("must be root to set number of bytes per line");
      }
    }
  }
#endif /* HAVE_LIBXXF86DGA */

  numbitplanes=QColor::numBitPlanes();

  if ( -1 == ioctl(fbttv, VIDIOCGMBUF, &(vmbuf)) ) 
    warning("screen: ioctl VIDIOCGMBUF");
  msize = vmbuf.size;

  // Alloc memory for snapshot 
  TargetMem=(char *)mmap(0,msize,PROT_READ|PROT_WRITE,MAP_SHARED,fbttv,0);
  if ((char*)-1 == TargetMem) {
    warning("screen: unable to allocate memory for snap shots and video clips!");
  }

#ifdef tvDEBUG
  debug("screen: visual %s, 0x%x, %d, dualmode %s",visname,myvisual,(long)visual_class,dual?"ON":"OFF");
  debug("screen: bpp %d, bpl %d, width %d, height %d",vbuf.depth,vbuf.bytesperline,vwin.width,vwin.height);
  debug("screen: video buffer: size %d, frames %d",vmbuf.size,vmbuf.frames);
  debug("screen: memory for snap shots: 0x%08x", (long) TargetMem);
  debug("screen: winId: 0x%x, parentWid: 0x%x",winId(),parentWid);
#endif

  colorMap8=false;
  // we have to use PseudoColor, i.e. we have only 256 colors
  if (visual_class == PseudoColor) {
    Colormap xcmap;
    int r, rr;
#ifdef tvDEBUG
    debug("screen: using color class PseudoColor");
#endif
    colorMap8=true;
    for (i=0;i<256;i++) {
      xcolors[i].pixel=i;
      xcolors[i].flags = DoRed | DoGreen | DoBlue;
    }
    //XQueryColors(disp, xcolormap, xcolors, 256);
    xcmap=XCreateColormap(disp, DefaultRootWindow(disp),myvisual, AllocAll);
    for (i=0;i<225;i++){ // Set Colors 16 - 240 
      r=i%25;
      xcolors[i+16].green =(65535*(i/25)/8);  
      xcolors[i+16].red =(65535*(r/5)/4);  
      rr=r%5;
      xcolors[i+16].blue =(65535*rr/4);        
    }
    XStoreColors(disp,xcmap,xcolors,256);
    //XSetWindowColormap(disp, (Window)WId(), xcmap);
    XSetWindowColormap(disp, (Window)parentWid, xcmap);
#ifdef tvDEBUG
    debug("screen: set up colormap for 8 bpp");
#endif
  }

  // set up some X stuff
  rootw=DefaultRootWindow(disp);

  // set XEvents we want to receive
  const uint stdWidgetEventMask =                 // X event mask
        KeyPressMask | KeyReleaseMask |
        ButtonPressMask | ButtonReleaseMask |
        KeymapStateMask |
        ButtonMotionMask |
        EnterWindowMask | LeaveWindowMask |
        FocusChangeMask |
        ExposureMask | VisibilityChangeMask |
        StructureNotifyMask | SubstructureNotifyMask;

  // tv window
  XSelectInput(disp,(Window) winId(), stdWidgetEventMask );
  //	       VisibilityChangeMask | StructureNotifyMask | SubstructureNotifyMask );

  // the shell window
  //XSelectInput(disp,parentWid, stdWidgetEventMask );

  // root window -- need visibility, sub-, structure
  XSelectInput(disp,rootw, 
	       VisibilityChangeMask | StructureNotifyMask | SubstructureNotifyMask );


  // init some variables
  allowViewRefresh=true;
  didViewRefresh=false;
  allowViewRefreshVisibility=true;
  didViewMove=false;
  viewRefreshDelay=DEFAULT_DELAY;
  captureWasOnQt=false;
  captureWasOnX11=false;
  infodelay=INFO_DELAY;

  saverState=1;
  cursorState=1;
  muteState=0;
  captureState=0;

  shotnumber=0;
  video_aktive=0;
  video_capturing=0;
  title_shown=0;

  setTVMaxSize();

  return(0);
}

//==============================================================================

void winTVScreenWidget::setTVMaxSize()
{
  uint max_h,max_w;
#if 1
  max_h=vcap.maxheight;
  max_w=vcap.maxwidth;
#else
  switch (defaultNorm) {
  case PAL:
  case SECAM:
    max_w=PAL_WIDTH;
    max_h=PAL_HEIGHT;
    break;

  case NTSC:
    max_w=NTSC_WIDTH;
    max_h=NTSC_HEIGHT;
    break;
    
  default:
    max_w=DEFAULT_WIDTH;
    max_h=DEFAULT_HEIGHT;
    break;
  }
#endif
  max_width=max_w;
  max_height=max_h;

#ifdef tvDEBUG
  debug ("screen: screensize max_x: %d, max_y: %d",max_width,max_height);
#endif
}

//==============================================================================

void winTVScreenWidget::destroyTV()
{
#ifdef tvDEBUG
  debug("screen: destroyTV");
#endif

  capture(0);

  if (!saverState) toggleScreenSaverSLOT();
    
#ifdef tvDEBUG
  debug("screen: reset XErrorHandler");
#endif
  XSetErrorHandler(oldXErrorHandler);

  if (colorMap8) {
    XFreeColormap(qt_xdisplay(),xcolormap);
  }

  XSelectInput(qt_xdisplay(),DefaultRootWindow(qt_xdisplay()),0);

#ifdef tvDEBUG
  debug("screen: freeing mem");
#endif
  if ((char*)-1 != TargetMem)
    munmap(TargetMem,msize); 
  
#ifdef tvDEBUG
  debug("screen: closing device %s",devname);
#endif
  if ( -1 == ::close(fbttv) ) 
    warning("kwintv: Error closing the Video4Linux device <%s>!",devname);
  
  free(vchan);
  free(vaudio);
  
#ifdef tvDEBUG
  debug("screen: destroyTV done");
#endif
}

//==============================================================================

void winTVScreenWidget::newTVGeometry( bool doViewRefresh )
{
  int xp,yp,the_width,the_height;
  int dx,dy;
  bool checkClipping;
  int oldx,oldy,oldw,oldh;

#ifdef tvDEBUG
  debug("screen: newTVGeometry");
#endif

  if (!isVisible()) return;

  if( -1 == ioctl(fbttv, VIDIOCGWIN, &vwin) )
    warning("screen: VIDIOCGWIN in ::newTVGeometry");

  oldx=vwin.x;
  oldy=vwin.y;
  oldw=vwin.width;
  oldh=vwin.height;
  vwin.flags=0;

  checkClipping=false;
  didViewMove=false;

  // get new geometry values, need the global position 
  QPoint p(0,0);
  p=mapToGlobal(p);
  xp=p.x();
  yp=p.y();
  the_height=geometry().height();
  the_width=geometry().width();
  //the_height=height();
  //the_width=width();
 
#ifdef tvDEBUG
  debug("screen: newTVGeometry: the_height %d, the_width %d",the_height,the_width);
#endif

  // checkClipping if size changed
  if(oldh!=the_height) checkClipping=true;
  if(oldw!=the_width) checkClipping=true;

  dx=0;dy=0;

#if 1
  // set a new width
  if (the_width > max_width) {
    dx=(the_width - max_width)/2;
    if (oldw != max_width) {
      checkClipping=true;
      vwin.width=max_width & ~3;
    }
  } else {
    vwin.width=the_width & ~3;
  }

  // set a new height
  if (the_height > max_height) {
    dy=(the_height - max_height)/2;
    if (oldh != max_height) {
      checkClipping=true;
      vwin.height=max_height & ~3;
    }
  } else {
    vwin.height=the_height & ~3;
  }
#else
  // set a new width
  if (the_width > max_width) {
    dx=(the_width - max_width)/2;
    if (oldw != max_width) {
      checkClipping=true;
      vwin.width=max_width;
    }
  } else {
    vwin.width=the_width;
  }

  // set a new height
  if (the_height > max_height) {
    dy=(the_height - max_height)/2;
    if (oldh != max_height) {
      checkClipping=true;
      vwin.height=max_height;
    }
  } else {
    vwin.height=the_height;
  }
#endif // #if 

  // checkClipping if position changed
  if (oldx!=xp+dx) {
    didViewMove=true;
    checkClipping=true;
  } else 
  if (oldy!=yp+dy) {
    didViewMove=true;
    checkClipping=true;
  } 

#if 1
  if ((vbuf.depth==24)&&((xp+dx)&3)) {
    xp=xp-((xp+dx) &3);
  }
  else if ((vbuf.depth==16)&&((xp+dx)&1)) {
    xp=xp-((xp+dx) &1);
  } 
  else if ((vbuf.depth==15)&&((xp+dx)&1)) {
      xp=xp-((xp+dx) &1);
  }

  if ((yp+dy)&1) {
    yp=yp-((yp+dy) &1);
  }
#endif

  if ( checkClipping ) {
    vwin.x=xp+dx;
    vwin.y=yp+dy;
    if (captureIsOn) {
#ifdef tvDEBUG
      debug("screen: newTVGeometry x %d, y %d , w %d, h %d, dx %d, dy %d",
	    vwin.x,vwin.y,vwin.width,vwin.height,dx,dy);
#endif // #if

      getTVClip(); 

      // this is done by getTVClip();
      //if ( -1 == ioctl(fbttv, VIDIOCSWIN, &vwin) )
      //fatal("screen: set clipped window");

      if ( allowViewRefresh && doViewRefresh ) refreshTV();
    }
  }  
}

//==============================================================================

void winTVScreenWidget::autoDetectAudioModeSLOT()
{
  vaudio[0].mode=0;
  if (-1 == ioctl(fbttv,VIDIOCGAUDIO,&vaudio[0]))
    warning("ioctl VIDIOCGAUDIO");
    //fatal("ioctl VIDIOCGAUDIO");
  if (audioMode != vaudio[0].mode) {
    //audioMode=vaudio[0].mode;
    emit autoDetectAudioModeSIGNAL(vaudio[0].mode | AUDIOMODE_AUTODETECT);
    showChannelInfoSLOT(NULL,CHINFO_NODISPVAL,0,0,audiomode[audioMode].str);
  }

#ifdef tvDEBUG
  //debug("screen: autoDetectAudioModeSLOT mode: %s",audiomode[vaudio[0].mode].str);
#endif
}

//==============================================================================

void winTVScreenWidget::refreshTVActionSLOT()
{
  int scr,depth;
  XSetWindowAttributes xswa;
  unsigned long mask;
  Window rwin,rootW;
  Display *disp;

#ifdef X11DEBUG
  debug("X11: refreshTVAction queried");
#endif

  viewRefreshTimer->stop();

  if ( allowViewRefresh ) {
#ifdef X11DEBUG
    debug("X11: do refreshTVAction");
#endif

    disp=qt_xdisplay();
    scr=DefaultScreen(disp);
    rootW= RootWindow(disp, scr);
    
    depth=numbitplanes;
    
    xswa.override_redirect = true;
    xswa.backing_store = NotUseful;
    xswa.save_under = false;
    mask = ( CWSaveUnder | CWBackingStore | CWOverrideRedirect );
    if (didViewMove) {
      rwin = XCreateWindow(disp, rootW, 0, 0, 
			   kapp->desktop()->width(),kapp->desktop()->height(),
			   /*vbuf.width, vbuf.height,*/
			   0, CopyFromParent, InputOutput, CopyFromParent,
			   mask, &xswa);
      didViewMove=false;
    } else {
      rwin = XCreateWindow(disp,rootW,
			   vwin.x-80, vwin.y-80, 
			   vwin.width+80,vwin.height+80,
			   0,CopyFromParent, InputOutput, CopyFromParent,
			   mask, &xswa);
    }
    //    lastrwin=rwin;
    XMapWindow(disp, rwin);
    XUnmapWindow(disp, rwin);
    XDestroyWindow(disp, rwin);
  }
  
  didViewRefresh=true;
}

//==============================================================================


void  winTVScreenWidget::refreshTV()
{
#ifdef X11DEBUG
  debug("X11: refreshTV");
#endif
  didViewRefresh=false;
  if ( viewRefreshTimer->isActive() ) viewRefreshTimer->stop();
  viewRefreshTimer->start(viewRefreshDelay);
}

//==============================================================================

void winTVScreenWidget::addTVClip(struct video_clip *cr,int *ncr,uint x,uint y,uint x2,uint y2)
{
  cr[*ncr].x=x;
  cr[*ncr].y=y;
  cr[*ncr].width=x2-x+1;
  cr[*ncr].height=y2-y+1;
  *ncr=*ncr+1;

#ifdef X11DEBUG
  debug("X11: addTVClip: added clip %d, %d, %d, %d",x,y,x2,y2);
#endif
} 

//==============================================================================

void winTVScreenWidget::makeTVClipTab( int ncr )
{
  vwin.clips=cliprecs;
  vwin.clipcount=ncr;
  
  if( -1 == ioctl(fbttv, VIDIOCSWIN, &vwin) )
    warning("screen: VIDIOCSWIN in ::makeTVClipTab");
}

//==============================================================================

int winTVScreenWidget::getTVClip(void)
{
  int ncr=0;
  int x,y,x2,y2,wx,wy;
  uint ww,wh;
  Display *disp;
  XWindowAttributes wts;
  Window parent,win,parent2, root, rroot, *children, swin;
  uint nchildren, i;
  struct video_clip *cr; 

  if (!isVisible()) return(0);

  disp=qt_xdisplay();

  win=(Window)winId();
  XQueryTree(disp,win, &rroot, &parent, &children, &nchildren);
  if(nchildren) XFree((char *)children);
 
  cr=cliprecs;

  /*
  if( -1 == ioctl(fbttv, VIDIOCGWIN, &vwin) )
    warning("screen: VIDIOCGWIN in ::getTVClip");
  */

  wx=vwin.x; wy=vwin.y;
  wh=vwin.height; ww=vwin.width;

#if 1
  if (wx<0)
    addTVClip(cr, &ncr, 0, 0, (uint)(-wx), wh-1);
  if (wy<0)
    addTVClip(cr, &ncr, 0, 0, ww-1, (uint)(-wy));
  if ((wx+ww) > vbuf.width)
    addTVClip(cr, &ncr, vbuf.width-wx, 0, ww-1, wh-1);
  if ((wy+wh) > vbuf.height)
    addTVClip(cr, &ncr, 0, vbuf.height-wy, ww-1, wh-1);
#endif
    
  root=DefaultRootWindow(disp);
  swin=win;
  while (parent!=root) {
    swin=parent;
    XQueryTree(disp, swin, &rroot, &parent, &children, &nchildren);
    if (nchildren)
      XFree((char *) children);
  }
  XQueryTree(disp, root, &rroot, &parent2, &children, &nchildren);
  for (i=0; i<nchildren;i++) 
    if (children[i]==swin) break;
  i++;

#ifdef X11DEBUG
  debug("X11: getTVClip: nchildren %d, %d",nchildren,i);
#endif

  for (; i<nchildren; i++) {
    
    XGetWindowAttributes(disp, children[i], &wts);
    if (!(wts.map_state)) continue;
    
    x=wts.x-vwin.x; 
    y=wts.y-vwin.y;
    //x2=x+wts.width+2*wts.border_width-1;
    //y2=y+wts.height+2*wts.border_width-1;
    x2=x+wts.width+2*wts.border_width;
    y2=y+wts.height+2*wts.border_width;

    if ((x2 < 0) || (x >= (int)ww) || (y2 < 0) || (y >= (int)wh)) continue;

    if (x<0) x=0;
    if (y<0) y=0;
    if (x2>=(int)ww) x2=ww-1;
    if (y2>=(int)wh) y2=wh-1;

    addTVClip(cr, &ncr, x, y, x2, y2);

  }
  XFree((char *) children);

  makeTVClipTab(ncr);

  return ncr;
}

//==============================================================================

void flip(char *mem, int len)
{
  int i;
  char tmp;

  for (i=0;i<len;i++) {
    tmp=mem[2];
    mem[2]=mem[0];
    mem[0]=tmp;
    mem+=3;
  }
}

//==============================================================================
//---------------------------------- SLOTS -------------------------------------
//==============================================================================

void winTVScreenWidget::snapShotSLOT(const char *prefix, int fixed, int width, int height, int format)
{
  //char name[555],ext[10];
  QString name,ext;
  char *tm=TargetMem;

  if ((tm != (char *)-1) && captureIsOn && !video_aktive) {

    if (fixed) {
      vgrab_even.width=width;
      vgrab_even.height=height;
    } else {
      vgrab_even.width=vwin.width;
      vgrab_even.height=vwin.height;
    }

    vgrab_even.format=VIDEO_PALETTE_RGB24;
    vgrab_even.frame=0;

    ioctl(fbttv, VIDIOCMCAPTURE, &vgrab_even);
    ioctl(fbttv, VIDIOCSYNC, &vgrab_even.frame);

    // dont overwrite existing files...
    switch (format) {
#ifdef HAVE_LIBJPEG
    case SNAPSHOT_JPEG:
      ext=".jpg";
      break;
#endif 
#ifdef HAVE_LIBGIF
    case SNAPSHOT_GIF:
      ext=".gif";
#endif
#ifdef HAVE_LIBTIFF
    case SNAPSHOT_TIFF:
      ext=".tif";
#endif
    case SNAPSHOT_PNM:
    default:
      ext=".pnm";
    }

    bool file_exists=true;
    while (file_exists) {
      QTextStream oname( name, IO_WriteOnly );
      oname << prefix << HEX(3) << shotnumber << ext << char(0);
      ifstream os_t(name);
      if (! os_t) {
	file_exists=false;
	os_t.close();
      } else {
	shotnumber++;
	os_t.close();
      }
    }

    QTextStream oname( name, IO_WriteOnly );
    oname << prefix << HEX(3) << shotnumber << ext << char(0);

    switch (format) {
#ifdef HAVE_LIBJPEG
    case SNAPSHOT_JPEG:
      {
	int rs,k;
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE *outfile; 

	if (!(outfile=fopen((const char*)name, "wb"))){
	  warning("screen: can't open file %s",(const char*)name);
	  return;
	}

	flip(tm, vgrab_even.width*vgrab_even.height);

	cinfo.err=jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_stdio_dest(&cinfo,outfile);
	cinfo.image_width=vgrab_even.width;
	cinfo.image_height=vgrab_even.height;
	cinfo.input_components=3;
	cinfo.in_color_space=JCS_RGB;
      
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo,80,true);
    
	jpeg_start_compress(&cinfo,true);
      
	rs=vgrab_even.width*3;
	JSAMPROW rp[vgrab_even.height];
	for (k=0; k<vgrab_even.height; k++)
	  rp[k]=(unsigned char *)&tm[rs*k];
	jpeg_write_scanlines(&cinfo,rp,vgrab_even.height);
      
	/* while(cinfo.next_scanline < cinfo.image_height){
	   JSAMPROW rp[1];
	   rp[0]=& mem[rs*cinfo.next_scanline];
	   int sl=jpeg_write_scanlines(&cinfo,rp,1);
	   }
	   */
	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	fclose(outfile);
      }
    break;
#endif 
    case SNAPSHOT_GIF:
    case SNAPSHOT_TIFF:
    case SNAPSHOT_PNM:
    default:
      {
	ofstream os((const char*)name);
	if (!os) {
	  warning("screen: can't open file %s!",(const char*)name);
	  return;
	}
	os << "P6\n" << dec << vgrab_even.width << "\n"
	   << dec << vgrab_even.height << "\n255\n";
      
	// FIX ME do we need this??
	flip(tm, vgrab_even.width*vgrab_even.height);
	os.write(tm,vgrab_even.width*vgrab_even.height*3);
	os.close();
	chown(name,getuid(),getgid());
      }
    }

#ifdef tvDEBUG
    debug("screen: snap shot %s",(const char*)name);
#endif

    showChannelInfoSLOT(NULL,CHINFO_NODISPVAL,0,0,(const char*)(QString("wrote ")+name));

  } // captureIsOn
}

//==============================================================================

long winTVScreenWidget::getChannelSignal()
{
  if ( ioctl(fbttv, VIDIOCGTUNER, &(vtuner)) == -1 )
    //fatal("screen: VIDIOCGTUNER");
    warning("screen: VIDIOCGTUNER");
#ifdef tvDEBUG
  debug("screen: signal quality: %d",vtuner.signal);
#endif
  return vtuner.signal;
}

//==============================================================================

void winTVScreenWidget::startVideoClipSLOT(struct MOVIE_PARAMS *movieparams)
{
#ifndef NOCARD
  if ((TargetMem != (char *)-1) && !video_aktive && captureIsOn) {
    video_aktive=1;

    switch (movieparams->video_clip) {
    case VIDEOCLIP_AVI: // avi
      video= new aviVideoClip();
      break;
    case VIDEOCLIP_PPM: // ppm
      video= new ppmVideoClip();
      break;
    case VIDEOCLIP_RAW: // raw
      video= new rawVideoClip();
      break;
    default:
      fatal("screen: unsupported video format %d",movieparams->video_clip);
    }

    // open video-file
    video->open(movieparams->filename,movieparams);

#ifdef tvDEBUG
    debug("screen: opened video-file %s",movieparams->filename);
#endif

    // set up vgrab
    vgrab_even.frame=0;
    vgrab_even.format=movieparams->video_format;
    vgrab_even.width=movieparams->width;
    vgrab_even.height=movieparams->height;

    memcpy(&vgrab_odd,&vgrab_even,sizeof(struct video_mmap));
    vgrab_odd.frame=1;

    video_frame=true;
    video_capturing=0;
    
    // los gehts
    writeVideoClipFrameSLOT();
  }
#endif
}

//==============================================================================
// looped version
void winTVScreenWidget::writeVideoClipFrameSLOT()
{
#ifndef NOCARD
  long lostframes=0;
  long frames=0;
  long grab,sync;
  int fps=video->parameter()->fps;
  int lfps=fps-1;
  bool lost=false;

  grab=0;sync=0;

  video_capturing=1;

  emit lostVideoClipFrameSIGNAL(lostframes);

  if (-1 == ioctl(fbttv,VIDIOCMCAPTURE,video_frame ? &vgrab_even : &vgrab_odd)) {
    if (errno == EAGAIN) 
      //debug(" %d", ioctl (fbttv, BTTV_PICNR,0));
      warning("EAGAIN");
    else
      warning("ioctl VIDIOCMCAPTURE start");
  }

  videoTime.start();

  while (video_aktive) {
    
    qApp->processEvents( 25 ); 

    elapsed=videoTime.elapsed();
    if ( (((frames * 1000.0) / (double)elapsed) <= fps) ) {

      video_frame=!video_frame;

      if (-1 == ioctl(fbttv,VIDIOCMCAPTURE,video_frame ? &vgrab_even : &vgrab_odd)) {
	if (errno == EAGAIN) 
	  //debug(" %d", ioctl (fbttv, BTTV_PICNR,0));
	  warning("EAGAIN");
	else
	  warning("ioctl VIDIOCMCAPTURE loop");
      } 

      if (-1 == ioctl(fbttv,VIDIOCSYNC,video_frame ? &vgrab_odd.frame : &vgrab_even.frame )) {
	if (errno != EAGAIN) 
	  warning("ioctl VIDIOCSYNC");
      } 

      //debug("%f",(frames * 1000.0) /(double)elapsed);
      video->writeframe((void*)(TargetMem + vmbuf.offsets[video_frame ? 0 : 1]));
      //video.writesound(...);
      if (!(video->nframes() % 5)) emit wroteVideoClipFrameSIGNAL(video->nframes());
      if (!(frames % 5)) emit wroteVideoClipFrameSIGNAL(frames);
      frames++;
      lost=false;
      elapsed=videoTime.elapsed();
      while ( ((frames * 1000.0) /(double)elapsed) < lfps ) {
	// lost a frame!
	lostframes++;
	frames++;
	lost=true;
	elapsed=videoTime.elapsed();
	debug("-");
      }
      if (lost && !(lostframes % 5)) emit lostVideoClipFrameSIGNAL(lostframes);

    } 
  }

  elapsed=videoTime.elapsed();

  video->close();

  frames=video->nframes();
  emit wroteVideoClipFrameSIGNAL(frames);
  emit statVideoClipFrameSIGNAL((frames*1000.0)/(double)elapsed);
  emit lostVideoClipFrameSIGNAL(lostframes);

  delete video;

#ifdef tvDEBUG
  
  debug("screen: wrote %d frames, %d msec. , %3.1f fps",frames,elapsed,(double)frames/elapsed*1000);
#endif
    
#endif // nocard
  video_capturing=0;
}
//processEvents ( int maxtime ) 
//==============================================================================

void winTVScreenWidget::endVideoClipSLOT()
{
#ifndef NOCARD
  if (video_aktive) {
    video_aktive=0;
#ifdef tvDEBUG
    debug("screen: video clip: stop");
#endif
  }
#endif // nocard
}

//==============================================================================

void winTVScreenWidget::changeFrequencySLOT(int value)
{
#ifndef NOCARD

  if ( defaultFreq != value ) {
    defaultFreq=value;
    if( -1 == ioctl(fbttv, VIDIOCSFREQ, &defaultFreq) )
      warning("screen: change frequency");
  }
  
#ifdef tvDEBUG
  debug("screen: changeFrequencySLOT %d",value);
#endif

#endif
}

//==============================================================================

void winTVScreenWidget::changeCaptureSLOT(int value)
{
  captureState=!captureState;
  /*  if (value) {
    captureIsOn=true;
    newTVGeometry();
  }*/
  capture(value);
}

//==============================================================================

void winTVScreenWidget::changeBrightnessSLOT(int bri)
{
#ifdef tvDEBUG
  debug("screen: changeBrightnessSLOT %d",bri);
#endif

#ifndef NOCARD
  struct video_picture vp;

  if (Bright != bri)
    {
      if(ioctl(fbttv, VIDIOCGPICT, &vp)==-1)
	warning("screen: get picture params");
      Bright=bri;
      vp.brightness=(Bright+128)<<8;
      if(ioctl(fbttv, VIDIOCSPICT, &vp))
	fatal("screen: set picture params");
      showChannelInfoSLOT(NULL,bri,-127,127,"Brightness");
    }
#endif // NOCARD
}

//==============================================================================

void winTVScreenWidget::changeContrastSLOT(int contr)
{
#ifdef tvDEBUG
  debug("screen: changeContrastSLOT %d",contr);
#endif

#ifndef NOCARD
  struct video_picture vp;

  if (Contrast != contr)
    {
      if(ioctl(fbttv, VIDIOCGPICT, &vp)==-1)
	warning("screen: get picture params");
      Contrast=contr;
      vp.contrast=Contrast<<7;
      if(ioctl(fbttv, VIDIOCSPICT, &vp))
	fatal("screen: set picture params");
      showChannelInfoSLOT(NULL,contr,0,511,"Contrast");
    }
#endif
}

//==============================================================================

void winTVScreenWidget::changeColorSLOT(int color)
{
#ifdef tvDEBUG
  debug("screen: changeColorSLOT %d",color);
#endif

#ifndef NOCARD
  struct video_picture vp;

  if (Color != color)
    {
      if(ioctl(fbttv, VIDIOCGPICT, &vp)==-1)
	warning("screen: get picture params");
      Color=color;
      vp.colour=Color<<7;
      if(ioctl(fbttv, VIDIOCSPICT, &vp))
	fatal("screen: set picture params");
      showChannelInfoSLOT(NULL,color,0,511,"Color");
    }
#endif
}

//==============================================================================

void winTVScreenWidget::changeHueSLOT(int hue)
{
#ifdef tvDEBUG
  debug("screen: changeHueSLOT %d",hue);
#endif

#ifndef NOCARD
  struct video_picture vp;

  if (Hue != hue)
    {
      if(ioctl(fbttv, VIDIOCGPICT, &vp)==-1)
	warning("screen: get picture params");
      Hue=hue;
      vp.hue=(Hue+128)<<8;
      if(ioctl(fbttv, VIDIOCSPICT, &vp))
	fatal("screen: set picture params");
      showChannelInfoSLOT(NULL,hue,-127,127,"Hue");
    }
#endif
}

//==============================================================================

void winTVScreenWidget::changeSourceSLOT(int s)
{
#ifdef tvDEBUG
  debug("screen: changeSourceSLOT %d",s);
#endif

#ifndef NOCARD
  if (defaultInput != s){
    defaultInput=s;
    if(ioctl(fbttv, VIDIOCSCHAN, &(defaultInput)))
      fatal("screen: change tuner");
  }
#endif
}

//==============================================================================

void winTVScreenWidget::changeNormSLOT(int n)
{
#ifdef tvDEBUG
  debug("screen: changeNormSLOT %d",n);
#endif

#ifndef NOCARD
  if (n != defaultNorm) {
    if (ioctl(fbttv, VIDIOCGTUNER, &vtuner))
      fatal("screen: ioctl VIDIOCGTUNER: change norm");
    vtuner.mode=n;
    defaultNorm=n;
    setTVMaxSize();
    newTVGeometry( true );
    if (ioctl(fbttv, VIDIOCSCHAN, &vtuner))
      fatal("screen: ioctl VIDIOCSTUNER: change norm");
  }
#endif
}

//==============================================================================

void winTVScreenWidget::changeChannelValuesSLOT(channel *c, int id)
{
  int pictmod=0;
  struct video_picture vp;

  if (!c) return;

#ifdef tvDEBUG
  debug("screen: change channel values: %s, %d",c->name,c->frequency);
#endif

#ifndef NOCARD
  if (captureIsOn) {
     switch_capture(0);
  }

  if (defaultFreq != c->frequency) {
    bool muted=false;
    if ( !muteState ) {
      toggleAudioMuteSLOT();
      muted=true;
    }

    defaultFreq=c->frequency;
    if(ioctl(fbttv, VIDIOCSFREQ, &defaultFreq))
      warning("screen: change frequency");

    if ( muted ) toggleAudioMuteSLOT();
  }
  
  /*
   *	Need to create a set format ioctl ??
   */

  if (c->norm != defaultNorm) {
    if (ioctl(fbttv, VIDIOCGTUNER, &vtuner))
      fatal("screen: ioctl VIDIOCGTUNER: change norm");
    vtuner.mode=c->norm;
    defaultNorm=c->norm;
    setTVMaxSize();
    newTVGeometry( true );
    if (ioctl(fbttv, VIDIOCSTUNER, &vtuner))
      fatal("screen: ioctl VIDIOCSTUNER: change norm");
  }

  
  if (defaultInput != c->input)
    {
      defaultInput=c->input;
      int v=defaultInput;
      if(ioctl(fbttv, VIDIOCSCHAN, &v))
	warning("screen: change input");
    }
  
  if( -1 == ioctl(fbttv, VIDIOCGPICT, &vp) )
    warning("screen: get picture params");
  
  if (Bright != c->brightness)
    {
      Bright=c->brightness;
      vp.brightness=(Bright+128)<<8;
      pictmod=1;
    }
  
  if (Color !=c->color)
    {
      Color=c->color;
      vp.colour=c->color<<7;
      pictmod=1;
    }
  
  if (Hue !=c->hue)
    {
      Hue=c->hue;
      vp.hue=(c->hue+128)<<8;
      pictmod=1;
    }
  
  
  if (Contrast !=c->contrast)
    {
      Contrast=c->contrast;
      vp.contrast=c->contrast<<7;
      pictmod=1;
    }
  
  if(pictmod)
    {
      if(ioctl(fbttv, VIDIOCSPICT, &vp))
	warning("screen: set picture params");
    }  

  switch_capture(1);

  c->id=id;
  showChannelInfoSLOT(c,CHINFO_NODISPVAL,0,0,NULL);

#ifdef tvDEBUG
  debug("screen: changed values");
#endif

#endif 
}

//==============================================================================

void winTVScreenWidget::toggleAudioMuteSLOT()
{
  muteState=!muteState;

#ifdef tvDEBUG
  debug("screen: changeMuteSLOT, muted=%s",muteState?"yes":"no");
#endif
  
#ifndef NOCARD
  if (muteState) {
    vaudio[0].flags |= VIDEO_AUDIO_MUTE;
    showChannelInfoSLOT(NULL,CHINFO_NODISPVAL,0,0,"Sound muted");
  } else {
    vaudio[0].flags &= ~VIDEO_AUDIO_MUTE;
    showChannelInfoSLOT(NULL,CHINFO_NODISPVAL,0,0,"Sound unmuted");
  }

  //vaudio[0].volume = audioVolume;

  if (-1 == ioctl(fbttv,VIDIOCSAUDIO,&vaudio[0]))
    fatal("ioctl VIDIOCSAUDIO");

#endif
}

//==============================================================================

void winTVScreenWidget::changeAudioVolumeSLOT(int vl, int vr)
{
#if 0

#ifdef tvDEBUG
    debug("screen: changeAudioVolumeSLOT set audio volume to %d, %d",vl,vr);
#endif
  if (!audioVolumeCan) return;

  if (muteState) vaudio[0].flags |= VIDEO_AUDIO_MUTE;
  else vaudio[0].flags &= ~VIDEO_AUDIO_MUTE;

  vaudio[0].volume = vr;
  vaudio[0].volume = vl;
  audioVolume=vl;

  //vaudio[0].mode = audioMode ? audioMode : 0;

    if (-1 == ioctl(fbttv,VIDIOCSAUDIO,&vaudio[0]))
	fatal("ioctl VIDIOCSAUDIO");
#endif   
}

//==============================================================================

void winTVScreenWidget::changeAudioModeSLOT(int mode)
{

  if ((mode > 0) && (mode < 5)) {
#ifdef tvDEBUG
    debug("screen: changeAudioModeSLOT set audio mode to %s",audiomode[mode].str);
#endif
    if (audioDetectTimer->isActive()) audioDetectTimer->stop();
    audioMode=audiomode[mode].nr;
    // try to set mode...
    vaudio[0].mode=audioMode;
    vaudio[0].volume = audioVolume;
    if (-1 == ioctl(fbttv,VIDIOCSAUDIO,&vaudio[0]))
      warning("ioctl VIDIOCSAUDIO");
    //fatal("ioctl VIDIOCGAUDIO");
    if (-1 == ioctl(fbttv,VIDIOCGAUDIO,&vaudio[0]))
      warning("ioctl VIDIOCGAUDIO");
    //fatal("ioctl VIDIOCGAUDIO");
    // set the real mode
    audioMode=vaudio[0].mode;
    // and inform others...
    emit autoDetectAudioModeSIGNAL(audioMode);
    showChannelInfoSLOT(NULL,CHINFO_NODISPVAL,0,0,audiomode[audioMode].str);
  } else {
    // autodetection, mode==0
    // set up timer
    audioMode=audiomode[mode].nr;
    if (audioDetectTimer->isActive()) audioDetectTimer->stop();
    audioDetectTimer->start(AUDIOMODE_DELAY);
  }

}

//==============================================================================

void winTVScreenWidget::changeTVMultiDisplaySLOT(int x, int y, int w, int h)
{
#ifdef tvDEBUG
  debug("screen: changeTVMultiDisplaySLOT %d, %d, %d, %d",x,y,w,h);
#endif
}

//==============================================================================

void winTVScreenWidget::toggleScreenSaverSLOT()
{
  int t,i,p,a;
  Display *displ=qt_xdisplay();

  saverState=!saverState;
  if (saverState) {
#ifdef tvDEBUG
  debug("screen: screensaver enabled; saverState = %d",saverState);
#endif
    XGetScreenSaver(displ,
		    &t,&i,&p,&a);
    if (t == 0 && i == ss_interval &&
	p == ss_prefer_blanking && a == ss_allow_exposures)
      XSetScreenSaver(displ,
		      ss_timeout,ss_interval,ss_prefer_blanking,ss_allow_exposures);
  } else {
#ifdef tvDEBUG
  debug("screen: screensaver disabled; saverState = %d",saverState);
#endif
    XGetScreenSaver(displ,
		    &ss_timeout,&ss_interval,&ss_prefer_blanking,&ss_allow_exposures);
    XSetScreenSaver(displ,
		    0,ss_interval,ss_prefer_blanking,ss_allow_exposures);
  }

}

//==============================================================================

void winTVScreenWidget::toggleMouseCursorSLOT()
{
  cursorState=!cursorState;

#ifdef tvDEBUG
  debug("screen: cursorState = %s",cursorState?"off":"on");
#endif

  if (!cursorState) {
    orgCursor=cursor();
    setCursor(blankCursor);
  } else {
    setCursor(orgCursor);
  }
}


//==============================================================================

void winTVScreenWidget::showChannelInfoEnableSLOT( int on )
{
  if (on) chInfo=AKTIVE;
  else chInfo=INAKTIVE;
}

//==============================================================================

void winTVScreenWidget::showChannelInfoSLOT(channel *ch,
					    int val, int min_val, int max_val, 
					    const char *text)
{
  
  if ( !isVisible() || (chInfo==INAKTIVE) ) return;

  chInfoLCD->setGeometry(vwin.x+vwin.width-(vwin.width/11)-20,vwin.y+20,
			 vwin.width/11,vwin.height/11);
  chInfoText->setGeometry(vwin.x+(vwin.width/10),vwin.y+vwin.height-(vwin.height/10)-10,
			  chInfoText->width(),vwin.height/10);
  chInfoBar->setGeometry(vwin.x+(vwin.width/10),vwin.y+vwin.height-(vwin.height/10)*2,
			 8*(vwin.width/10),vwin.height/20);
  
  if (ch) {
    chInfoBar->hide();
    if (!chInfoLCD->isVisible()) chInfoLCD->show();
    chInfoLCD->display(ch->id);
    chInfoLCD->raise();

    if (!chInfoText->isVisible()) chInfoText->show();
    chInfoText->setText(ch->name);
    chInfoText->raise();

    chInfoText->repaint();
    chInfoLCD->repaint();
  }

  if ( text ) {
    chInfoLCD->hide();
    chInfoBar->hide();
    if (!chInfoText->isVisible()) chInfoText->show();
    chInfoText->setText(text);
    chInfoText->raise();
    chInfoText->repaint();
  }

  if ( val != CHINFO_NODISPVAL ) {
    chInfoLCD->hide();
    if (! text ) chInfoText->hide();
    if (!chInfoBar->isVisible()) chInfoBar->show();
    chInfoBar->setRange(min_val,max_val);
    chInfoBar->setValue(val);
    chInfoBar->raise();
    chInfoBar->repaint();
  }

  if (infoTimer->isActive()) infoTimer->stop();
  infoTimer->start(infodelay,true);
#ifdef tvDEBUG
  debug("screen: showChannelInfoSLOT");
#endif
}

//==============================================================================

void winTVScreenWidget::unShowChannelInfoSLOT()
{
  if (chInfoLCD->isVisible()) {
    chInfoLCD->hide();
  }
  if (chInfoText->isVisible()) {
    chInfoText->hide();
  }
  if (chInfoBar->isVisible()) 
    chInfoBar->hide();

  infoTimer->stop();
}

//==============================================================================

void winTVScreenWidget::readData()
{
  if( -1 == ioctl(fbttv, VIDIOCGWIN, &vwin) )
    warning("screen: VIDIOCGWIN in ::readData");
}

int winTVScreenWidget::getTVX()
{ 
  readData();
  return vwin.x;
}

int winTVScreenWidget::getTVY()
{ 
  readData();
  return vwin.y;
}

int winTVScreenWidget::getTVHeight()
{ 
  readData();
  return vwin.height;
}

int winTVScreenWidget::getTVWidth()
{ 
  readData();
  return vwin.width;
}

//==============================================================================
//==============================================================================
// never change the next line!
#include "wintvscreen.moc"

// end of file









